"""base.py: Versions of code in simple.py that are extendable in ways necessary for
some applications

"""

from __future__ import annotations  # Enables, eg, (self: HMM,

import typing  # For type hints

import numpy
import numpy.random

import hmm.simple


class Observation_0:
    """Base class for observations.  You can't use instances of this
    class.  You must use a subclass.

    """
    _parameter_keys: typing.List[str] = ['n_states']

    # Specifies parameters reported by self.__str__.  These parameter
    # names should be in self.__dict__.

    def __init__(self: Observation_0, *args):
        self._rng = args[-1]
        self.n_states = 0  # Subclasses must set n_states
        for key in self._parameter_keys:
            assert key in self.__dict__

    def _concatenate(self: Observation_0, y_segs: typing.Sequence) -> tuple:
        """Concatenate observation segments.  Assumes each is a numpy array.

        Args:
            y_segs:  The observations.  The structure depends on the subclass

        Returns: (tuple): (y, t_seg) where the structure of y depends
            on the subclass, and t_seg is a list of segment
            boundaries.

        """
        assert isinstance(y_segs, (tuple, list))
        length = 0
        t_seg = [0]
        for seg in y_segs:
            length += len(seg)
            t_seg.append(length)
        try:
            all_data = numpy.concatenate(y_segs)
        except:
            print('Observation_0._concatenate() cannot handle type {0}'.format(
                type(y_segs[0])))
            raise
        return all_data, t_seg

    def reestimate(self: Observation_0, w: numpy.ndarray):
        """Based on w, modify parameters of self

        Args:
            w (numpy.ndarray):  A statistic of all the data, y, and old
                parameters of HMM including self.
                w[t, s] = Prob(state[t]=s|y and HMM)

        Sets parameters of self to \argmax_\theta Prob(y|w) (Todo:
        verify)

        """
        raise NotImplementedError('Not implemented.  Use a subclass.')

    def random_out(self: Observation_0, state: int):
        """ Returns a random draw from Prob(y|state)

        Args:
            state: The state

        Returns:
            (object): A single observation

        """
        raise NotImplementedError('Not implemented.  Use a subclass.')

    def calculate(self: Observation_0) -> numpy.ndarray:
        """Calculate and return the likelihoods of states given observations.

        Returns:
            self._likelihood where _likelihood[t,i] = Prob(y[t]|state[t]=i)

        self.observe must be called first to assign measured data to self.
        """
        raise NotImplementedError('Not implemented.  Use a subclass.')

    def merge(self: Observation_0, raw_outs) -> numpy.ndarray:
        """Reformat raw_outs into suitable form for self.observe()

        Motivating example: raw_outs is a list of tuples (bundle, y)
        from Observation_with_bundles.random_out().

        """
        return numpy.array(raw_outs)

    def observe(self: Observation_0, y_segs) -> numpy.ndarray:
        """Attach observations to self

        Args:
            y_segs: Independent measurement sequences.  Structure
                specified by implementation of self._concatenate() by
                subclasses.

        Returns:
            t_segs: Segment boundaries

        """
        # pylint: disable = attribute-defined-outside-init
        self._y, t_seg = self._concatenate(y_segs)
        self.t_seg = numpy.array(t_seg)
        self.n_times = t_seg[-1]
        self._likelihood = numpy.empty((self.n_times, self.n_states))
        return self.t_seg

    def __str__(self: Observation_0) -> str:
        """Return a string representation of self.

        Returns:
            A string representation of self

        """
        return_string = 'An {0} instance:\n'.format(type(self))
        for key in self._parameter_keys:
            return_string += '    {0}\n'.format(key)
            return_string += '{0}\n'.format(getattr(self, key))
        return return_string


class IntegerObservation(Observation_0):
    r"""Observation model for integers with y[t] \in [0:y_max] \forall t

    Args:
        py_state: Conditional probability of each y give each state
        rng: A numpy.random.Generator for simulation

    """

    _parameter_keys = ['_py_state']

    def __init__(self: IntegerObservation, py_state: numpy.ndarray,
                 rng: numpy.random.Generator):
        self._py_state = hmm.simple.Prob(py_state)
        super().__init__(rng)

        self._cummulative_y = numpy.cumsum(self._py_state, axis=1)
        self.n_states = len(self._py_state)

    def reestimate(  # pylint: disable = arguments-differ
            self: IntegerObservation,
            w: numpy.ndarray,
            warn: typing.Optional[bool] = True):
        """
        Estimate new model parameters.

        Args:
            w: w[t,s] = Prob(state[t]=s) given data and
                 old model
            warn: If True and y[0].dtype != numpy.int32, print
                warning
        """
        if not (isinstance(self._y, numpy.ndarray) and
                (self._y.dtype == numpy.int32)):
            # pylint: disable = attribute-defined-outside-init
            self._y = numpy.array(self._y, numpy.int32)
            if warn:
                print("Warning: reformatted y in reestimate")
        assert self._y.dtype == numpy.int32 and self._y.shape == (
            self.n_times,), """
                y.dtype=%s, y.shape=%s""" % (
                self._y.dtype,
                self._y.shape,
            )
        for yi in range(self._py_state.shape[1]):
            self._py_state.assign_col(
                yi,
                w.take(numpy.where(self._y == yi)[0], axis=0).sum(axis=0))
        self._py_state.normalize()
        self._cummulative_y = numpy.cumsum(self._py_state, axis=1)

    def calculate(self: IntegerObservation) -> numpy.ndarray:
        r"""
        Calculate likelihoods: self._likelihood[t,i] = P(y(t)|state(t)=i)

        Returns:
            state_likelihood[t,i] \forall t \in [0:n_times] and i \in [0:n_states]

        Assumes a previous call to measure has assigned self._y and allocated
            self._likelihood

        """

        # mypy objects ""Unsupported target for indexed assignment"
        self._likelihood[:, :] = self._py_state[:, self._y].T  # type: ignore
        return self._likelihood

    def random_out(self: IntegerObservation, state: int) -> int:
        """For simulation, draw a random observation given state s

        Args:
            state: Index of state

        Returns: Random observation drawn from distribution
            conditioned on the state

        """
        return numpy.searchsorted(self._cummulative_y[state],
                                  self._rng.random())


class HMM(hmm.simple.HMM):
    """Uses observations that are in many segments.  Also supports
    decoding sequences of bundles rather than states if the
    observation class has information about bundles.

    """
    # Allow y_mod type to conflict with hmm.simple.HMM
    y_mod: Observation_0  # type: ignore

    def reestimate(self: HMM):
        """Phase of Baum Welch training that reestimates model parameters

        Using values af self.alpha and self.beta calculated by
        forward() and backward(), this code updates state transition
        probabilities and initial state probabilities.  The call to
        y_mod.reestimate() updates observation model parameters.

        """

        # u_sum[i,j] = \sum_{t:gamma_inv[t+1]>0} alpha[t,i] * beta[t+1,j] *
        # state_likelihood[t+1,j]/gamma[t+1]
        #
        # The term at t is the conditional probability that there was
        # a transition from state i to state j at time t given all of
        # the observed data
        u_sum = numpy.einsum(
            "ti,tj,tj,t->ij",  # Specifies the i,j indices and sum over t
            self.alpha[:-1],  # indices t,i
            self.beta[1:],  # indices t,j
            self.state_likelihood[1:],  # indices t,j
            self.gamma_inv[1:]  # index t
        )
        # Make corrections for terms on segment boundaries
        for t in numpy.nonzero(self.gamma_inv < 0)[0]:
            assert self.gamma_inv[t] < 0
            if t == 0:
                continue
            for i in range(self.n_states):
                for j in range(self.n_states):
                    u_sum[i, j] -= self.alpha[t - 1, i] * self.beta[
                        t, j] * self.state_likelihood[t, j] * self.gamma_inv[t]

        self.alpha *= self.beta  # Saves allocating a new array for
        alpha_beta = self.alpha  # the result

        self.p_state_time_average = alpha_beta.sum(axis=0)  # type: ignore
        self.p_state_initial = numpy.copy(alpha_beta[0])
        for x in (self.p_state_time_average, self.p_state_initial):
            x /= x.sum()
        assert u_sum.shape == self.p_state2state.shape
        self.p_state2state *= u_sum
        self.p_state2state.normalize()
        self.y_mod.reestimate(alpha_beta)

    def forward(  # pylint: disable = arguments-differ
            self: HMM,
            t_start: int = 0,
            t_stop: int = 0,
            t_skip: int = 0,
            last_0=None) -> float:
        """Recursively calculate state probabilities, P(s[t]|y[0:t])

        Args:
            t_start: Use self.state_likelihood[t_start] first
            t_stop: Use self.state_likelihood[t_start] first
            t_skip: Number of time steps from when "last" is valid till t_start

        Returns:
            Log (base e) likelihood of HMM given entire observation sequence

        Works on a single uninterrupted sequence of observations that
        may be only part of a larger set.  As side effect, it sets
        self.alpha and self.gamma_inv for the same range of t as the
        observations.

        """
        if t_stop == 0:
            # Reduces to ignoring t_start and t_stop and operating on
            # a single segment
            assert t_start == 0
            t_stop = len(self.state_likelihood)

        if last_0 is None:
            last_0 = self.p_state_initial

        last = numpy.copy(last_0).reshape(-1)
        for t in range(t_skip):
            self.p_state2state.step_forward(last)

        for t in range(t_start, t_stop):
            last *= self.state_likelihood[t]  # Element-wise multiply
            assert last.sum() > 0
            self.gamma_inv[t] = 1 / last.sum()
            last *= self.gamma_inv[t]
            self.alpha[t, :] = last
            self.p_state2state.step_forward(last)
        return -(numpy.log(self.gamma_inv[t_start:t_stop])).sum()

    def backward(  # pylint: disable = arguments-differ
            self: HMM, t_start=0, t_stop=0):
        """
        Baum Welch backwards pass through state conditional likelihoods.


        Calculates values of self.beta which "reestimate()" needs.
        """

        if t_stop == 0:
            # Reduces to ignoring t_start and t_stop and operating on
            # a single segment
            assert t_start == 0
            t_stop = len(self.state_likelihood)

        last = numpy.ones(self.n_states)
        for t in range(t_stop - 1, t_start - 1, -1):
            self.beta[t, :] = last
            last *= self.state_likelihood[t] * self.gamma_inv[t]
            self.p_state2state.step_back(last)

    def multi_train(self: HMM, ys, n_iterations: int, display=True):
        """Train on more than one independent sequence of observations

        Args:
            ys: Measured observation sequences in format appropriate
                for self.y_mod
            n_iter: The number of iterations to execute

        Returns:
            List of utility (log likelihood or log MAP) per observation
            for each iteration

        Generalizes simple.HMM by allowing more than one independent
        observation sequence.  Also accommodates MAP estimation
        instead of MLE.

        For the first training iteration, the initial distribution of
        states at the beginning of each observation sequence is given
        by self.p_state_initial.  However, each training iteration
        updates the distribution of the initial state for each
        observation sequence independently.  At the end of training,
        the average of those independent distributions for the initial
        states is assigned to self.p_state_initial.

        I don't know of assumptions for which that procedure is
        exactly correct.  If one knows that each observation segment
        starts with the same state distribution, then one should use a
        single initial state distribution in training.  If one knows
        that the initial state distributions are different and that
        when the model is used it will be applied to new data drawn
        from similar sets of segments, then one should train with
        separate initial state distributions and retain them for use
        in the future.

        The compromise treatment of the initial distribution of states
        here is intended to treat initial distributions that are
        different from each other but more similar to each other than
        the time averaged state distribution.

        """

        # pylint: disable = attribute-defined-outside-init

        # utility_list[i] = log(Prob(ys|HMM[iteration=i]))/n_times, ie,
        # the log likelihood per time step, or MAP
        utility_list = []

        t_seg = self.y_mod.observe(ys)  # Segment boundaries
        self.n_times = self.y_mod.n_times

        assert self.n_times > 1
        assert t_seg[0] == 0

        n_seg = len(t_seg) - 1
        self.alpha = numpy.empty((self.n_times, self.n_states))
        self.beta = numpy.empty((self.n_times, self.n_states))
        self.gamma_inv = numpy.empty((self.n_times,))

        # State probabilities at the beginning of each segment
        p_state_initial_all = numpy.empty((n_seg, self.n_states))
        for seg in range(n_seg):
            p_state_initial_all[seg, :] = self.p_state_initial.copy()

        for iteration in range(n_iterations):
            message = "i={0:4d} ".format(iteration)
            sum_log_like = 0.0
            self.state_likelihood = self.y_mod.calculate()

            # Operate on each observation segment separately and put
            # the results in the corresponding segement of the alpha,
            # beta and gamma_inv arrays.

            for seg in range(n_seg):
                # Set up self to run forward and backward on this segment
                t_start = t_seg[seg]
                t_stop = t_seg[seg + 1]

                log_likelihood = self.forward(
                    t_start, t_stop, last_0=p_state_initial_all[seg, :])
                self.backward(t_start, t_stop)

                sum_log_like += log_likelihood
                p_state_initial_all[
                    seg, :] = self.alpha[t_start] * self.beta[t_start]
                # Flag to prevent fitting state transitions between segments
                self.gamma_inv[t_start] = -1
                message += "L[{0}]={1:7.4f} ".format(
                    seg, log_likelihood / self.n_times)

            # Record/report/check this iteration
            if hasattr(self.y_mod, 'log_prior'):
                utility = sum_log_like + self.y_mod.log_prior()  # type: ignore
            else:
                utility = sum_log_like
            utility_list.append(utility / self.n_times)
            message += "avg={0:10.7f}".format(utility_list[-1])
            self.ensure_monotonic(utility_list, display, message)

            self.reestimate()

        self.p_state_initial[:] = p_state_initial_all.sum(axis=0)
        self.p_state_initial /= self.p_state_initial.sum()
        return utility_list

    def initialize_y_model(
            self: HMM,
            y,
            state_sequence: typing.Optional[numpy.ndarray] = None):
        """Given data, make a plausible y_model.

        Args:
            y: Observation sequence.  Type must work for self.y_mod.observe(y)
            state_sequence: Optional state sequence

        Use this method to create an observation model that is not the
        same for every state and that makes the data in the argument y
        plausible.

        """
        # ToDo: Fails for general y_mod
        n_times = self.y_mod.observe(y)[-1]
        if state_sequence is None:
            state_sequence = numpy.array(self.state_simulate(n_times),
                                         numpy.int32)

        # Set alpha to enforce the simulated state sequence in reestimate
        alpha = numpy.zeros((n_times, self.n_states))
        for t in range(n_times):
            alpha[t, state_sequence[t]] = 1

        self.y_mod.reestimate(alpha)
        return self.y_mod

    def simulate(self: HMM, length: int):
        """Simulate n steps of HMM

        Args:
            length: Number of time steps to simulate

        Returns:
            (states, outs) where states is a state sequence and outs
            is a sequence of observations

        """

        states, raw_outs = super().simulate(length)
        # y_mod.merge enables funny observation models like
        # Observation_with_bundles
        return states, self.y_mod.merge(raw_outs)


class Bundle_segment:
    """A pair of time series: bundle ids and observations.

    Args:
        bundles: Tags for each time
        y: Observations for each time

    """

    def __init__(self: Bundle_segment, bundles, y):
        self.bundles = bundles
        self.y = y

    def __len__(self: Bundle_segment) -> int:
        return len(self.bundles)

    def __str__(self: Bundle_segment) -> str:
        return """y values:{0}
bundle values:{1}
""".format(self.y, self.bundles)


class Observation_with_bundles(Observation_0):
    """Represents likelihood of states given observations by masking an
    underlying likelihood model.

    Args:
        underlying_instance: An instance of the underlying model class
        bundle2state: Keys are bundle ids and values are lists of states
        rng: A numpy.random.Generator for simulation
        small: Threshold for plausible sum of state likelihoods at a time

    :ivar underlying_model:  Provides methods for observations without bundle tags
    :ivar bundle2state: A dict with bundle2state[i_bundle] = list of states
    :ivar n_bundle:
    :ivar n_states:
    :ivar bundle_and_state: A boolean array with state in bundle <=>
                            bundle_and_state[i_bundle, i_state] = True
    :ivar state2bundle: i_bundle = state2bundle[i_state] with state in bundle
    """
    _parameter_keys = 'underlying_model bundle2state'.split()

    def __init__(self: Observation_with_bundles,
                 underlying_instance: Observation_0,
                 bundle2state: dict,
                 rng: numpy.random.Generator,
                 small=1.0e-30):
        self.underlying_model = underlying_instance
        assert isinstance(self.underlying_model, Observation_0)
        self.bundle2state = bundle2state
        self.n_bundle = len(self.bundle2state)
        self.small = small
        # Call super().__init__ to check all _parameter_keys assigned
        super().__init__(rng)

        # Find out how many states there are and ensure that each
        # state is in only one bundle.
        states: typing.Set[int] = set()
        for x in self.bundle2state.values():
            # Assert that no state is in more than one bundle.  Think
            # before relaxing this.
            assert states & set(x) == set()
            states = states | set(x)
        self.n_states = len(states)
        # Ensure that states is a set of sequential integers [0,n_states)
        assert states == set(range(self.n_states))

        # bundle_and_state[bundle_id, state_id] is true iff state \in bundle.
        self.bundle_and_state = numpy.zeros((self.n_bundle, self.n_states),
                                            bool)

        # state2bundle[state_id] = bundle_id for bundle that contains state
        self.state2bundle = numpy.ones(self.n_states, dtype=numpy.int32) * -1

        for bundle_id, states in self.bundle2state.items():
            for state_id in states:
                self.bundle_and_state[bundle_id, state_id] = True
                # Ensure that no state is in more than one bundle
                assert self.state2bundle[state_id] == -1
                self.state2bundle[state_id] = bundle_id

    def observe(  # pylint: disable = arguments-differ
            self: Observation_with_bundles,
            bundle_segment_list: list) -> numpy.ndarray:
        """Attach observations to self as a single Bundle_segment

        Args:
            bundle_segment_list: List of Bundle_segments

        Side effects: 1. Assign self.t_seg; 2. Call self.y_mod.observe
        to attach observations stripped of bundle tags to self.y_mod.

        """
        # pylint: disable = attribute-defined-outside-init
        self._y, self.t_seg = self._concatenate(  # type: ignore
            bundle_segment_list)
        self.n_times = self.t_seg[-1]
        self.underlying_model.observe([self._y.y])
        return self.t_seg

    def _concatenate(  # pylint: disable = arguments-differ
        self: Observation_with_bundles,
        bundle_segment_list: typing.Sequence) -> typing.Tuple[Bundle_segment,
                                                              list]:
        """ Create a single Bundle_segment from a list of segments

        Args:
            bundle_segment_list: Each element is a Bundle_segment

        Returns:
            (y, t_seg)
        """
        length = 0
        t_seg = [0]
        bundles = []
        ys = []
        for segment in bundle_segment_list:
            length += len(segment.bundles)
            t_seg.append(length)
            bundles.append(segment.bundles)
            ys.append(segment.y)

        all_bundles = [item for sublist in bundles for item in sublist]
        all_ys, y_t_seg = self.underlying_model._concatenate(ys)
        assert tuple(t_seg) == tuple(y_t_seg)
        return Bundle_segment(all_bundles, all_ys), t_seg

    def random_out(self: Observation_with_bundles, state: int) -> tuple:
        """ Draw a single output from the conditional distribution given state.
        """
        return self.state2bundle[state], self.underlying_model.random_out(state)

    def merge(  # type: ignore
            self: Observation_with_bundles, raw_outs: list) -> Bundle_segment:
        """ Merge isolated pairs (bundle, y) into an Observation_with_bundles.

        Args:
            raw_outs: A list of pairs (bundle[t], y[t])

        Returns:
            A single Observation_with_bundles instance

        """
        bundles = []
        y = []
        for out in raw_outs:
            bundles.append(out[0])
            y.append(out[1])

        return Bundle_segment(numpy.array(bundles),
                              self.underlying_model.merge(y))

    def calculate(self: Observation_with_bundles) -> numpy.ndarray:
        """Calculate and return likelihoods of states given self._y.

        Returns:
            likelihood with likelihood[t,s] = Probability(y[t]|state[t]=s) *
                Probability(state=s|bundle[t])

        Assumes self._y has been assigned by a call to self.observe().
        For each time t, the given the observation (bundle, y) return
        value is a vector of length n_states with components
        Probability(y|state)*Probability(state|bundle).

        """

        # pylint: disable = attribute-defined-outside-init
        # Get unmasked likelihoods
        self._likelihood = self.underlying_model.calculate()

        # Apply the right mask for the bundle at each time.  Note this
        # modifies self.y_mod._likelihood in place.
        for t in range(self.n_times):
            self._likelihood[t, :] *= self.bundle_and_state[
                self._y.bundles[t], :]
            if self._likelihood[t, :].sum() < self.small:
                raise ValueError(
                    'Observation is not plausible from any state.  ' +
                    'self.likelihood[{0},:]=\n{1}'.format(
                        t, self._likelihood[t, :]))

        return self._likelihood

    def reestimate(self: Observation_with_bundles, w: numpy.ndarray):
        """Reestimate parameters of self.y_mod

        Args:
            w: Weights with w[t,s] = alpha[t,s]*beta[t,s] =
                Probability(state=s|all data)

        Assumes that observations are already attached to self.y_mod by
        self.observe().

        """
        self.underlying_model.reestimate(w)



""" simple.py: Implements basic HMM algorithms.

Classes:

    :py:class:`HMM`:
        A Hidden Markov Model implementation

    :py:class:`Observation`:
        Models of discrete observations

    :py:class:`Prob`:
        A subclass of ndarray for conditional probability matrices

"""
# Nomenclature:
#
# y:                Observations
#
# y_mod:            An instance of an observation model, eg, Observation in this file
#
# n_times:          The number of time points in data for an observable
#
# state_likelihood: Given observed data y[t] = y_ and states[t] = s_,
#                   state_likelihood[t, s_] = Prob(y_ | s_)
#

# pylint: disable = attribute-defined-outside-init
from __future__ import annotations  # Enables, eg, (self: HMM,

import typing  # For type hints

import numpy
import numpy.random

COPYRIGHT = """Copyright (c) 2021 Andrew M. Fraser

This file is part of hmm.

Hmm is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

Hmm is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

See the file gpl.txt in the root directory of the hmm distribution
or see <http://www.gnu.org/licenses/>.
"""


class HMM:
    """A Hidden Markov Model implementation.

    Args:
        p_state_initial : Initial distribution of states
        p_state_time_average : Time average distribution of states
        p_state2state : Probability of state given state:
            p_state2state[a, b] = Prob(s[1]=b|s[0]=a)
        y_mod : Instance of class for probabilities of observations
        rng : Numpy generator with state

    p_state_time_average is averaged over training data.  It is not
    the stationary distribution of p_state2state.

    Arguments are passed by reference and they are modified by some of
    the methods of HMM.

    By initializing with rng created by the caller with, eg
    numpy.random.default_rng(seed), one can ensure reproducible
    pseudo-random sequences and avoid using the same pseudo-random
    sequences in different parts of the code.

    """

    def __init__(self: HMM,
                 p_state_initial: numpy.ndarray,
                 p_state_time_average: numpy.ndarray,
                 p_state2state: numpy.ndarray,
                 y_mod: Observation,
                 rng: typing.Optional[numpy.random.Generator] = None) -> None:

        if rng is None:
            self.rng = numpy.random.default_rng()
        else:
            self.rng = rng
        self.n_states = len(p_state_initial)
        self.p_state_initial = numpy.array(p_state_initial)
        self.p_state_time_average = numpy.array(p_state_time_average)
        self.p_state2state = Prob(numpy.array(p_state2state))
        self.y_mod = y_mod
        self.gamma_inv: numpy.ndarray = None
        self.alpha: numpy.ndarray = None
        self.beta: numpy.ndarray = None

    def forward(self: HMM) -> float:
        """Recursively calculate state probabilities.

        Returns:
            Log (base e) of likelihood of HMM given entire observation sequence

        Requires that observation probabilities have already been calculated

        On entry:

        - self                    is an HMM

        - self.state_likelihood   has been calculated

        - self.alpha              has been allocated

        - self.gamma_inv          has been allocated

        On return:

        - 1/self.gamma_inv[t] = Prob{y[t]=y[t]|y[0:t]}
        - self.alpha[t, i] = Prob{s[t]=i|y[0:t+1]}

        """

        # last is a conditional distribution of state probabilities.
        # What it is conditioned on changes as the calculations
        # progress.
        last = numpy.copy(self.p_state_initial.reshape(-1))  # Copy
        for t in range(len(self.state_likelihood)):
            last *= self.state_likelihood[t]  # Element-wise multiply
            assert last.sum() > 0  # Fails if data is impossible
            self.gamma_inv[t] = 1 / last.sum()
            last *= self.gamma_inv[t]
            self.alpha[t, :] = last
            last[:] = numpy.dot(last, self.p_state2state)
        return -(numpy.log(self.gamma_inv)).sum()

    def backward(self: HMM) -> None:
        """
        Baum Welch backwards pass through state conditional likelihoods.


        Calculates values of self.beta which "reestimate()" needs.

        On entry :

        - self               is an HMM

        - self.state_likelihood  has been calculated

        - self.gamma_inv     has been calculated by forward

        - self.beta          has been allocated

        On return:

        - For each state i, beta[t, i] = Prob(y[t+1:T]|s[t]=i)/Prob(y[t+1:T])

        """
        # last and beta are analogous to last and alpha in forward(),
        # but the precise interpretations are more complicated.
        last = numpy.ones(self.n_states)
        for t in range(len(self.state_likelihood) - 1, -1, -1):
            self.beta[t, :] = last
            last *= self.state_likelihood[t] * self.gamma_inv[t]
            last[:] = numpy.dot(self.p_state2state, last)

    def calculate(self: HMM):
        """Prepares for calling forward if it is not called from train.

        """
        self.state_likelihood = self.y_mod.calculate()
        self.n_times = self.y_mod.n_times
        self.alpha = numpy.empty((self.n_times, self.n_states))
        self.gamma_inv = numpy.empty((self.n_times,))

    def train(
            self: HMM,
            y,  #  Type must work for self.y_mod.observe(y)
            n_iterations: int = 1,
            display: typing.Optional[bool] = True) -> list:
        """Use Baum-Welch algorithm to search for maximum likelihood
        model parameters.

        Args:
            y: Measurements appropriate for self.y_mod
            n_iter: The number of iterations to execute
            display: If True, print the log likelihood
                per observation for each iteration

        Returns:
            List of log likelihood per observation for each iteration

        """

        log_likelihood_list = []
        # Attach observations to self.y_mod
        self.y_mod.observe(y)
        self.n_times = self.y_mod.n_times
        assert self.n_times > 1

        # Allocate working arrays
        self.alpha = numpy.empty((self.n_times, self.n_states))
        self.beta = numpy.empty((self.n_times, self.n_states))
        self.gamma_inv = numpy.empty((self.n_times,))

        for iteration in range(n_iterations):
            self.state_likelihood = self.y_mod.calculate()
            log_likelihood = self.forward()
            self.backward()
            self.reestimate()

            log_likelihood_list.append(log_likelihood / self.n_times)
            self.ensure_monotonic(
                log_likelihood_list, display,
                "{0:4d}: LLps={1:7.3f}".format(iteration,
                                               log_likelihood_list[-1]))

        return log_likelihood_list

    def ensure_monotonic(self: HMM, log_likelihood_list, display, message):
        if display:
            print(message)
        if len(log_likelihood_list) == 1:
            return

        ll = log_likelihood_list[-1]
        ll_prev = log_likelihood_list[-2]
        delta = ll - ll_prev
        if delta / abs(ll) < -1.0e-15:  # Todo: Why not zero?
            iteration = len(log_likelihood_list)
            raise ValueError("""
Training is not monotonic: LLps[{0}]={1} and LLps[{2}]={3} difference={4}
""".format(iteration - 1, ll_prev, iteration, ll, delta))

    def reestimate(self: HMM):
        """Phase of Baum Welch training that reestimates model parameters

        Using values af self.alpha and self.beta calculated by
        forward() and backward(), this code updates state transition
        probabilities and initial state probabilities.  The call to
        y_mod.reestimate() updates observation model parameters.

        """

        # u_sum[i,j] = \sum_t alpha[t,i] * beta[t+1,j] *
        # state_likelihood[t+1]/gamma[t+1]
        #
        # The term at t is the conditional probability that there was
        # a transition from state i to state j at time t given all of
        # the observed data
        u_sum = numpy.einsum(
            "ti,tj,tj,t->ij",  # Specifies the i,j indices and sum over t
            self.alpha[:-1],  # indices t,i
            self.beta[1:],  # indices t,j
            self.state_likelihood[1:],  # indices t,j
            self.gamma_inv[1:]  # index t
        )
        self.alpha *= self.beta  # Saves allocating a new array for
        alpha_beta = self.alpha  # the result

        self.p_state_time_average = alpha_beta.sum(axis=0)  # type: ignore
        self.p_state_initial = numpy.copy(alpha_beta[0])
        for x in (self.p_state_time_average, self.p_state_initial):
            x /= x.sum()
        assert u_sum.shape == self.p_state2state.shape
        self.p_state2state *= u_sum
        self.p_state2state.normalize()
        self.y_mod.reestimate(alpha_beta)

    def decode(self: HMM, y) -> numpy.ndarray:
        """
        Find the most likely state sequence for given observation sequence.

        Args:
            y: Observations with type for self.y_mod or None if
                self.state_likelihood was assigned externally.

        Returns:
            Maximum likelihood state sequence

        This implements the Viterbi algorithm.
        """

        if y is not None:  # Calculate likelihood of data given state
            self.y_mod.observe(y)
            self.n_times = self.y_mod.n_times
            self.state_likelihood = self.y_mod.calculate()
        n_times, n_states = self.state_likelihood.shape
        assert self.n_states == n_states
        assert n_times > 1

        # Allocate working memory
        best_predecessors = numpy.empty((self.n_times, self.n_states),
                                        numpy.int32)
        best_state_sequence = numpy.ones((self.n_times, 1), numpy.int32)

        # Use initial state distribution for first best_path_utility
        best_path_utility = self.state_likelihood[0] * self.p_state_initial

        for t in range(1, self.n_times):
            # utility = p_state2state*outer(best_path_utility, state_likelihood[t])
            utility = (self.p_state2state.T *
                       best_path_utility).T * self.state_likelihood[t]
            best_predecessors[t] = utility.argmax(axis=0)
            best_path_utility = numpy.choose(best_predecessors[t], utility)
            if best_path_utility.max() == 0:
                raise ValueError(
                    "Attempt to decode impossible observation sequence")
            best_path_utility /= best_path_utility.max()  # Prevent underflow

        # Find the best end state
        previous_best_state = numpy.argmax(best_path_utility)

        # Backtrack through best_predecessors to find the best
        # sequence.
        for t in range(self.n_times - 1, -1, -1):
            best_state_sequence[t] = previous_best_state
            previous_best_state = best_predecessors[t, previous_best_state]
        return best_state_sequence.reshape(-1)

    def state_simulate(
        self: HMM,
        length: int,
        mask: typing.Optional[numpy.ndarray] = None,
    ) -> numpy.ndarray:
        """Generate a random sequence of states that is perhaps constrained
        by a mask.

        Args:
            length: Length of returned array

        Keyword Args:
            mask: If mask[t, i] is False, state i is forbidden at time t.

        Returns:
            Sequence of states

        The returned sequence is not a draw from the random process
        defined by the model.  However the sequence has probability >
        0.  The function does not use an observation function and does
        not simulate observations.  The simulate method provides draws
        of states and observations from a fully specified model.

        """

        self.state_likelihood = self.rng.random((length, self.n_states))
        self.n_times = length
        if mask is not None:
            self.state_likelihood *= mask

        try:
            state_sequence = self.decode(None)
        except ValueError as e:
            raise Exception(
                "State_simulate given an impossible mask constraint") from e

        return state_sequence

    def simulate(
        self: HMM,
        length: int,
    ) -> typing.Tuple[numpy.ndarray, numpy.ndarray]:
        """Generate a random draw from the model of sequences of states and
        observations of a given length.

        Args:
            length: Number of time steps to simulate

        Returns:
            (states, outs) where states[t] is the state at time t, and
                outs[t] is the output at time t.

        """

        # Initialize lists
        outs = []
        states = []
        # Set up cumulative distributions
        cumulative_initial = numpy.cumsum(self.p_state_time_average[0])
        cumulative_transition = numpy.cumsum(self.p_state2state.values(),
                                             axis=1)

        # cum_rand generates random integers from a cumulative distribution
        def cum_rand(cum):
            return numpy.searchsorted(cum, self.rng.random())

        # Select an initial state
        state = cum_rand(cumulative_initial)
        # Select subsequent states and call model to generate observations
        for _ in range(length):
            states.append(state)
            outs.append(self.y_mod.random_out(state))
            state = cum_rand(cumulative_transition[state])
        return numpy.array(states), numpy.array(outs)

    def link(self: HMM, here: int, there: int, p: float):
        """Create (or remove) a link between state "here" and state "there".

        Args:
            here: One index of element to modify
            there: Another index of element to modify
            p: Weight or probability of link

        The strength of the link is a function of both the argument
        "p" and the existing conditional probabilities of state
        transitions, self.p_state2state, in which
        self.p_state2state[here, there] is the probability of going to
        state there given that the system is in state here.  The code
        sets p_state2state[here, there] to p and then re-normalizes
        self.p_state2state.  Set self.p_state2state itself if you need
        to set exact values.  You can use this method to modify the
        state topology by removing links before training.

        """
        self.p_state2state[here, there] = p
        self.p_state2state[here, :] /= self.p_state2state[here, :].sum()

    def __str__(self: HMM) -> str:  # HMM instance
        return """{0} with {1:d} states
p_state_initial:      {2}
p_state_time_average: {3}
p_state2state =
{4}
{5}""".format(
            self.__class__,
            self.n_states,
            self.p_state_initial,
            self.p_state_time_average,
            self.p_state2state.values(),
            self.y_mod,
        )

    def deallocate(self: HMM) -> HMM:
        """ Remove arrays assigned by train.

        To be called before writing a model to disk
        """
        del (self.alpha)
        del (self.beta)
        del (self.gamma_inv)
        return self


class Observation:
    """ Probability models for observations drawn from a set of sequential integers.

    Args:
        py_state:  Conditional probability of y given state
        rng: A numpy.random.Generator for simulation

    Public methods and attributes:

    __init__

    observe

    random_out

    calculate

    reestimate

    n_states

    """

    def __init__(self: Observation,
                 py_state: numpy.ndarray,
                 rng: numpy.random.Generator = None):
        self._py_state = Prob(py_state)
        if rng is None:
            self._rng = numpy.random.default_rng()
        else:
            self._rng = rng
        self._cummulative_y = numpy.cumsum(self._py_state, axis=1)
        self.n_states = len(self._py_state)

    def observe(self: Observation, y) -> int:
        """ Attach measurement sequence[s] to self.

        Args:
            y: A sequence of integer observations

        Returns:
            Length of observation sequence
        """
        self._y = y
        self.n_times = len(self._y)

        # Allocate here rather than in calculate() because calculate()
        # may be called more often than observe().
        self._likelihood = numpy.empty((self.n_times, self.n_states),
                                       dtype=numpy.float64)
        return self.n_times

    def calculate(self: Observation) -> numpy.ndarray:
        r"""
        Calculate likelihoods: self._likelihood[t,s] = P(y[t]|state[t]=s)

        Returns:
            state_likelihood[t,s] \forall t \in [0:n_times] and s \in [0:n_states]

        Assumes a previous call to observe has assigned self._y and allocated
            self._likelihood

        """

        # mypy objects ""Unsupported target for indexed assignment"
        self._likelihood[:, :] = self._py_state[:, self._y].T  # type: ignore
        return self._likelihood

    def random_out(self: Observation, state: int) -> int:
        """For simulation, draw a random observation given state s

        Args:
            state: Index of state

        Returns: Random observation drawn from distribution
            conditioned on the state

        """
        return numpy.searchsorted(self._cummulative_y[state],
                                  self._rng.random())

    def reestimate(self: Observation, w: numpy.ndarray):
        """
        Estimate new _py_state

        Args:
            w: w[t,s] = Prob(state[t]=s) given data and old model
        """

        # Loop over range of allowed values of y
        for yi in range(self._py_state.shape[1]):
            # yi was observed at times: numpy.where(self._y == yi)[0]
            # w.take(...) is the conditional state probabilities at those times
            self._py_state.assign_col(
                yi,
                w.take(numpy.where(self._y == yi)[0], axis=0).sum(axis=0))
        self._py_state.normalize()
        self._cummulative_y = numpy.cumsum(self._py_state, axis=1)


class Prob(numpy.ndarray):
    """Subclass of ndarray for conditional probability matrices.  P[a,b]
    is the probability of b given a.  The class has additional methods
    and is designed to enable alternative implementations that run
    faster or in less memory but may be implemented by uglier code.

    """

    def __new__(cls, x: numpy.ndarray):
        """ Return a Prob instance of the argument.

        Args:
            x: An array of conditional probabilities

        """
        assert len(x.shape) == 2
        # cls is Prob.  This calls __new__ of numpy.ndarray and makes
        # the return value a Prob instance.
        return super().__new__(cls, x.shape, buffer=x.data)

    # See http://docs.scipy.org/doc/numpy/user/basics.subclassing.html
    def normalize(self: Prob) -> Prob:  # Prob instance
        """
        Modify self to make each row sum to one

        Returns:
            Self after normalization

        """
        s = self.sum(axis=1)
        for i in range(self.shape[0]):
            self[i, :] /= s[i]
        return self

    def assign_col(self: Prob, i: int, col: numpy.ndarray):
        """
        Replace a column of self with data specified by the arguments

        Args:
            i: Column index
            col: Column value

        Returns:
            Self after assignment
        """
        self[:, i] = col
        return self

    def likelihoods(self: Prob, v: numpy.ndarray) -> numpy.ndarray:
        r"""Likelihoods for vector of data

        Args:
            v: A time series of integer observations

        Returns:
            2-d array of state likelihoods

        If self represents probability of observing integers given
        state, ie, self[s, y] = Probability(observation=y|state=s),
        then this function returns the likelihood for each state given
        the observation at a particular time.  Given T = len(v) and
        self.shape = (M,N), this returns L with L.shape = (T,M) and L[t,a] =
        Prob(v[t]|a) \forall t \in [0:T] and a in [0:M].

        """
        return self[:, v].T

    def utility(self: Prob, nu: numpy.ndarray, py: numpy.ndarray):
        """Efficient calculation of numpy.outer(nu, py)*self (where * is
        element-wise)

        Args:
            nu:  Utility of maximum utility path to each state
            py: Likelihood of each state given data y[t]
        Returns:
            Maximum utilities for sequences ending in state pairs

        Used in Viterbi decoding with self[a,b] =
        Prob(s[t+1]=b|s[t]=a).  If nu[a] = maximum utility of s[t-1]=a
        given the data y[0:t] and py[b] = Probability observation =
        y[t] given s[t]=b, then this method returns a 2-d array, C,
        with C[a,b] = utility of maximum utility path ending with s[t-1]=a,
        s[t]=b given observations y[0:t+1].

        """
        return (self.T * nu).T * py

    def step_forward(self: Prob, alpha: numpy.ndarray):
        """Replace values of argument a with matrix product a*self.

        Args:
            alpha (numpy.ndarray):  Alpha[t]

        Used in forward algorithm.  In the vector argument
        alpha[a]=Probability(s[t]=a|y[0:t+1]).  The resulting value is a
        vector A with A[a] = Probability(s[t+1]=a|y[0:t+1]).

        Not done inline because function written in c is better

        """
        alpha[:] = numpy.dot(alpha, self)

    def step_back(self: Prob, b: numpy.ndarray):
        """Replace values of argument a with matrix product self*a

        Args:
            b: See b[t] in the book

        Used in backward algorithm.  The vector result is beta[t-1]
        which is sort of like the vector alpha in step_forward.  See
        Chapter 2 of the book for a precise explanation.  The
        argument, b, already includes the probability of the
        observation at time t.  The calculation here applies the
        conditional state probability matrix backwards.

        Not done inline because function written in c is better

        """
        b[:] = numpy.dot(self, b)

    def values(self: Prob) -> Prob:
        """
        Produce values of self

        Returns:
            self

        This is a hack to free subclasses from the requirement of self
        being an nd_array

        """
        return self


# --------------------------------
# Local Variables:
# mode: python
# End:



"""test_observe_float.py: T run "$ python -m pytest test_observe_float.py" or
"$ python -m pytest hmm/tests"

"""
# Copyright (c) 2021 Andrew M. Fraser
import unittest

import numpy
import numpy.testing

import scipy.linalg  # type: ignore

import hmm.observe_float
import hmm.base


class Parent(unittest.TestCase):
    """ Don't use this class.  Use subclasses
    """

    # If you use method names for this class that begin with test,
    # they will be called with an instance of class Parent even though
    # the name Parent doesn't begin with Test.

    def setUp(self):

        p_initial_state = [0.67, 0.33]
        p_state2state = [[0.93, 0.07], [0.13, 0.87]]
        self.rng = numpy.random.default_rng(0)
        y_mod_a, y_mod_b = self.make_y_mods()  # pylint: disable = no-member
        self.model_a = hmm.base.HMM(p_initial_state, p_initial_state,
                                    p_state2state, y_mod_a, self.rng)
        self.model_b = hmm.base.HMM(p_initial_state, p_initial_state,
                                    p_state2state, y_mod_b)

        if hasattr(self.model_a.y_mod, 'initialize_out'):
            self.model_a.y_mod.initialize_out()  # For AutoRegressive y_mod
        _, y_train = self.model_a.simulate(100)
        self.y_shape = y_train.shape
        self.y_type = y_train.dtype

        self.y_train = numpy.array(y_train, numpy.float64)

    def _decode(self, sequence_sum):
        """ Called by test_decode in subclasses
        """
        rv = numpy.array(self.model_a.decode((self.y_train,)))
        self.assertTrue(rv.sum() == sequence_sum)

    def _train(self):
        """ Called by test_train in subclasses
        """
        self.model_b.y_mod.observe((self.y_train,))

        # Exercises calculate and reestimate
        rv = numpy.array(
            self.model_b.multi_train((self.y_train,), n_iterations=100))

        difference = rv[1:] - rv[:-1]
        self.assertTrue(difference.min() > -1.0e-15)  # Check monotonic

    def _str(self, reference_string):
        """ Called by test_str in subclasses
        """
        string = self.model_a.y_mod.__str__()
        n_instance = string.find('instance')
        tail = string[n_instance:]
        self.assertTrue(tail == reference_string)


class TestGauss(Parent):

    def make_y_mods(self):
        mu_1 = numpy.array([-1.0, 1.0])
        var_1 = numpy.ones(2)
        y_mod_a = hmm.observe_float.Gauss(mu_1.copy(), var_1.copy(), self.rng)
        y_mod_b = hmm.observe_float.Gauss(mu_1 * 2, var_1 * 4, self.rng)
        return y_mod_a, y_mod_b

    def test_decode(self):
        self._decode(49)

    def test_train(self):
        self._train()

    def test_str(self):
        self._str('instance:\n    mu\n[-1.  1.]\n    variance\n[1. 1.]\n')


class TestGaussMAP(TestGauss):

    def make_y_mods(self):
        mu_1 = numpy.array([-1.0, 1.0])
        var_1 = numpy.ones(2)
        y_mod_a = hmm.observe_float.GaussMAP(mu_1.copy(), var_1.copy(),
                                             self.rng)
        y_mod_b = hmm.observe_float.GaussMAP(mu_1 * 2, var_1 * 4, self.rng)
        return y_mod_a, y_mod_b


class TestMultivariate(Parent):

    def make_y_mods(self):
        mu_1 = numpy.array([[1.0, 1.0], [-1.0, -1.0]])
        mu_2 = numpy.array([[1.0, -1.0], [-1.0, 1.0]])
        var = numpy.array([[[1.0, 0.0], [0.0, 1.0]], [[1.0, 0.0], [0.0, 1.0]]])
        y_mod_a = hmm.observe_float.MultivariateGaussian(
            mu_1.copy(), var.copy(), self.rng)
        y_mod_b = hmm.observe_float.MultivariateGaussian(
            mu_2.copy(), var.copy(), self.rng)
        return y_mod_a, y_mod_b

    def test_decode(self):
        self._decode(47)

    def test_train(self):
        self._train()

    def test_str(self):
        self._str(
            'instance\nFor state 0:\n inverse_sigma = \n[[1. 0.]\n [0. 1.]]\n mu = [1. 1.] norm = 0.159155\nFor state 1:\n inverse_sigma = \n[[1. 0.]\n [0. 1.]]\n mu = [-1. -1.] norm = 0.159155\n'  # pylint: disable = line-too-long
        )


class TestAutoRegressive(Parent):
    """ AR order: 1
    """

    def make_y_mods(self):
        ar_1 = numpy.array([[0.5], [-0.5]])
        offset_1 = numpy.array([0.5, -0.5])
        ar_2 = numpy.array([[0.85], [-0.85]])
        offset_2 = numpy.array([-0.5, 0.5])
        var = numpy.array([1.0, 1.0])

        y_mod_a = hmm.observe_float.AutoRegressive(ar_1.copy(), offset_1.copy(),
                                                   var.copy(), self.rng)

        y_mod_b = hmm.observe_float.AutoRegressive(ar_2.copy(), offset_2.copy(),
                                                   var.copy(), self.rng)

        return y_mod_a, y_mod_b

    def test_decode(self):
        self._decode(42)

    def test_train(self):
        self._train()

    def test_str(self):
        self._str(
            'instance\nFor state 0:\n variance = \n1.0\n ar_coefficients = [0.5] offset = 0.5 norm = 0.398942\nFor state 1:\n variance = \n1.0\n ar_coefficients = [-0.5] offset = -0.5 norm = 0.398942\n'  # pylint: disable = line-too-long
        )


# --------------------------------
# Local Variables:
# mode: python
# End:



"""test_simple.py Tests hmm.simple

hmm.simple.Observation is tested with hmm.base.IntegerObservation in
test_base.py

$ python -m pytest hmm/tests/test_simple.py

"""

import unittest

import numpy
import numpy.testing
import numpy.random

import scipy.linalg  # type: ignore

import hmm.simple


class BaseClass(unittest.TestCase):
    """Holds common values and methods used by other test classes.

    Because this class has no method names that start with "test",
    unittest will not discover and run any of the methods.  However
    the methods of this class can be assigned to names that start with
    "test" in subclasses where they will be discovered and run.

    """
    # Values here can be accessed by instances of subclasses as, eg,
    # self.n_states.

    n_states = 6
    n_times = 1000
    p_state_initial = numpy.ones(n_states) / float(n_states)
    p_state_time_average = p_state_initial
    p_state2state = scipy.linalg.circulant([0, 0, 0, 0, 0.5, 0.5])
    _py_state = scipy.linalg.circulant([0.4, 0, 0, 0, 0.3, 0.3])

    # Because of the decorator, setUpClass is called with the class,
    # not an instance as the first argument, hence it is customarily
    # named cls.  It is called once while the method named setUp gets
    # called before each test.
    @classmethod
    def setUpClass(cls):
        """ This code is executed once per test class
        """
        # cls._py_state = scipy.linalg.circulant([0.4, 0, 0, 0, 0.3, 0.3])
        cls.mask = numpy.ones((cls.n_times, cls.n_states), bool)
        for t in range(cls.n_times):
            cls.mask[t, t % cls.n_states] = False

    def initialize(self):
        # pylint: disable = attribute-defined-outside-init
        self.rng = numpy.random.default_rng(0)
        self.simple_observation = hmm.simple.Observation(
            self._py_state.copy(), self.rng)
        self.simple_hmm = hmm.simple.HMM(
            self.p_state_initial.copy(),  # Initial distribution of states
            self.p_state_time_average.copy(
            ),  # Stationary distribution of states
            self.p_state2state.copy(),  # State transition probabilities
            self.simple_observation,
            rng=self.rng,
        )


class TestHMM(BaseClass):
    """ Test base.HMM
    """

    def setUp(self):
        super().initialize()
        self.mods = (self.simple_hmm,
                    )  # More mods after getting models in C built
        self.s, y = self.simple_hmm.simulate(self.n_times)
        self.y = numpy.array(y, numpy.int32).reshape((-1))

    def test_state_simulate(self):
        result1 = self.simple_hmm.state_simulate(self.n_times)
        result2 = self.simple_hmm.state_simulate(self.n_times, self.mask)
        for result in (result1, result2):
            self.assertTrue(len(result) == self.n_times)
            array = numpy.array(result)
            self.assertTrue(array.min() == 0)
            self.assertTrue(array.max() == self.n_states - 1)

    def test_link(self):
        """ Remove link from 0 to itself
        """
        self.simple_hmm.link(0, 0, 0)
        self.assertTrue(self.simple_hmm.p_state2state[0, 0] == 0.0)

    def test_str(self):
        string = self.simple_hmm.__str__()
        self.assertTrue(string.find('with 6 states') == 25)

    def test_unseeded(self):
        """ Exercise initialization of HMM without supplying rng
        """

        mod = hmm.simple.HMM(
            self.p_state_initial.copy(),  # Initial distribution of states
            self.p_state_time_average.copy(
            ),  # Stationary distribution of states
            self.p_state2state.copy(),  # State transition probabilities
            self.simple_observation,
        )
        states, y = mod.simulate(10)  # pylint: disable = unused-variable
        self.assertTrue(len(states) == 10)

    def test_decode(self):
        """
        Check that self.mod gets 70% of the states right
        """
        states = self.simple_hmm.decode(self.y)
        wrong = numpy.where(states != self.s)[0]
        self.assertTrue(len(wrong) < len(self.s) * .3)
        # Check that other models get the same state sequence as self.hmm
        for mod in self.mods[1:]:
            wrong = numpy.where(states != mod.decode(self.y))[0]
            self.assertTrue(len(wrong) == 0)

    def test_train(self):
        """ Test training
        """
        log_like = self.simple_hmm.train(self.y, n_iterations=10, display=True)
        # Check that log likelihood increases montonically
        for i in range(1, len(log_like)):
            self.assertTrue(log_like[i - 1] < log_like[i])
        # Check that trained model is close to true model
        numpy.testing.assert_allclose(self.simple_hmm.y_mod._py_state.values(),
                                      self._py_state,
                                      atol=0.15)
        numpy.testing.assert_allclose(self.simple_hmm.p_state2state.values(),
                                      self.p_state2state,
                                      atol=0.2)
        # Check that other models give results close to self.hmm
        for mod in self.mods[1:]:
            log_like_mod = mod.train(self.y, n_iter=10, display=False)
            numpy.testing.assert_allclose(log_like_mod, log_like)
            numpy.testing.assert_allclose(
                mod.y_mod._py_state.values(),
                self.simple_hmm.y_mod._py_state.values())
            numpy.testing.assert_allclose(
                mod.p_state2state.values(),
                self.simple_hmm.p_state2state.values())


class TestProb(BaseClass):
    """ Tests hmm.simple.Prob
    """

    A_ = numpy.array([[0, 2, 2.0], [2, 2, 4.0], [6, 2, 2.0]])
    B_ = numpy.array([[0, 1], [1, 1], [1, 3.0]])
    C_ = numpy.array([[0, 0, 2.0], [0, 0, 1.0], [6, 0, 0.0]])

    def setUp(self):
        """ This code is run before each test
        """
        self.a = hmm.simple.Prob(self.A_.copy())
        self.b = hmm.simple.Prob(self.B_.copy())
        self.c = hmm.simple.Prob(self.C_.copy())
        self.ms = (self.a, self.b, self.c)
        for m in self.ms:
            m.normalize()

    def test_normalize(self):
        for m in self.ms:
            n_rows, n_columns = m.shape
            for i in range(n_rows):
                s = 0
                for j in range(n_columns):
                    s += m.values()[i, j]
                numpy.testing.assert_almost_equal(1, s)

    def test_assign(self):
        a = self.c.values().sum()
        self.c.assign_col(1, [1, 1, 1])
        numpy.testing.assert_almost_equal(self.c.values().sum(), a + 3)

    def test_likelihoods(self):
        numpy.testing.assert_allclose(
            self.c.likelihoods([0, 1, 2])[2], [1, 1, 0])

    def test_utility(self):
        numpy.testing.assert_almost_equal(
            self.c.utility(self.b.T[0], self.b.T[1]),
            [[0, 0, 0], [0, 0, 0.375], [0.25, 0, 0]])

    def test_step_forward(self):
        b = numpy.array([1., 0.5, 0.75])
        self.a.step_forward(b)
        numpy.testing.assert_almost_equal(b, [0.575, 0.775, 0.9])

    def test_step_back(self):
        b = numpy.array([1., 0.5, 0.75])
        self.a.step_back(b)
        numpy.testing.assert_almost_equal(b, [0.625, 0.75, 0.85])

    def test_values(self):
        numpy.testing.assert_almost_equal(self.c.values(),
                                          [[0, 0, 1], [0, 0, 1], [1, 0, 0]])


# --------------------------------
# Local Variables:
# mode: python
# End:



"""test_base.py Tests hmm.base

$ python -m pytest hmm/tests/test_base.py or $ py.test --pdb hmm/tests/test_base.py

"""

import unittest

import numpy
import numpy.testing
import numpy.random

import scipy.linalg  # type: ignore

import hmm.base
import hmm.simple
import tests.test_simple


class TestObservations(unittest.TestCase):
    """ Test Observation in modules simple and base
    """

    def setUp(self):
        self.numpy_rng = numpy.random.default_rng(0)  # 0 is seed
        p_ys = hmm.simple.Prob(numpy.array([[0, 1], [1, 1], [1, 3.0]]))
        p_ys.normalize()
        n = 20
        y = numpy.empty(n, dtype=numpy.int32)
        for i in range(n):
            y[i] = (i + i % 2 + i % 3 + i % 5) % 2
        self.y = y
        self.y64 = numpy.array(y, dtype=numpy.int64)
        self.w = numpy.array(20 * [0, 0, 1.0]).reshape((n, 3))
        self.w[0, :] = [1, 0, 0]
        self.w[3, :] = [0, 1, 0]
        self.ys = (y[5:], y[3:7], y[:4])
        self.y_mod_base = hmm.base.IntegerObservation(p_ys.copy(),
                                                      self.numpy_rng)
        self.y_mod_simple = hmm.simple.Observation(p_ys.copy(), self.numpy_rng)
        # discrete model and observations from base
        self.y_mod_y_base = (self.y_mod_base, (self.y64,))
        # discrete model and observations from simple
        self.y_mod_y_simple = (self.y_mod_simple, self.y64)

    def calc(self, y_mod, y):
        y_mod.observe(y)
        py = y_mod.calculate()[2:4]
        numpy.testing.assert_almost_equal(py, [[0, 0.5, 0.25], [1, 0.5, 0.75]])

    def test_calc(self):
        for y_mod, y in (self.y_mod_y_base, self.y_mod_y_simple):
            self.calc(y_mod, y)

    def test_join(self):
        self.y_mod_base.observe(self.ys)
        numpy.testing.assert_equal(self.y_mod_base.t_seg, [0, 15, 19, 23])

    def reestimate(self, y_mod, y):
        y_mod.observe(y)
        y_mod.calculate()
        y_mod.reestimate(self.w)
        numpy.testing.assert_almost_equal([[1, 0], [0, 1], [5 / 9, 4 / 9]],
                                          y_mod._py_state.values())

    def test_reestimate(self):
        for y_mod, y in (self.y_mod_y_base, self.y_mod_y_simple):
            self.reestimate(y_mod, y)

    def test_str(self):
        self.assertTrue(isinstance(self.y_mod_base.__str__(), str))


class BaseClass(tests.test_simple.BaseClass):
    bundle2state = {0: [0, 1, 2], 1: [3], 2: [4, 5]}


class TestHMM(BaseClass):
    """ Test hmm.base.HMM
    """

    def setUp(self):
        self.y_class = hmm.base.IntegerObservation
        self.p_ys = hmm.simple.Prob(self._py_state.copy())
        self.rng = numpy.random.default_rng(0)

        # Setup for test_initialize_y_model with simple_observation
        simple_observation = hmm.base.IntegerObservation(self.p_ys, self.rng)
        self.simple_hmm = hmm.base.HMM(
            self.p_state_initial.copy(),  # Initial distribution of states
            self.p_state_initial.copy(),  # Stationary distribution of states
            self.p_state2state.copy(),  # State transition probabilities
            simple_observation,
            rng=self.rng,
        )
        _, observations = self.simple_hmm.simulate(1000)
        self.simple_y = [observations] * 5

        # Setup for other tests with bundle_observation
        bundle_observation = hmm.base.Observation_with_bundles(
            simple_observation, self.bundle2state, self.rng)
        self.base_hmm = hmm.base.HMM(
            self.p_state_initial.copy(),  # Initial distribution of states
            self.p_state_initial.copy(),  # Stationary distribution of states
            self.p_state2state.copy(),  # State transition probabilities
            bundle_observation,
            rng=self.rng,
        )
        _, observations = self.base_hmm.simulate(1000)
        self.y = [observations] * 5

    def test_initialize_y_model(self):
        """ Also exercises self.mod.state_simulate.
        """
        # Need .copy() because initialize_y_model modifies _py_state
        difference = self.simple_hmm.y_mod._py_state.copy(
        ) - self.simple_hmm.initialize_y_model(self.simple_y)._py_state
        self.assertTrue(difference.max() > 0.01)
        zero = difference.sum(axis=1)
        self.assertTrue(zero.max() < 1e-9)  # Rows of each should sum to one

    def test_state_simulate(self):
        self.base_hmm.state_simulate(10)

    def test_simulate(self):
        n = 10
        result = self.base_hmm.simulate(n)
        self.assertTrue(len(result[0]) == n)
        self.assertTrue(len(result[1].bundles) == n)

    def test_str(self):
        self.assertTrue(isinstance(self.base_hmm.__str__(), str))

    def test_t_skip(self):
        self.base_hmm.y_mod.observe(self.y[:1])
        self.base_hmm.state_likelihood = self.base_hmm.y_mod.calculate()
        n_times = self.base_hmm.y_mod.n_times
        n_states = self.base_hmm.n_states
        self.base_hmm.alpha = numpy.empty((n_times, n_states))
        self.base_hmm.gamma_inv = numpy.empty((n_times,))
        self.assertTrue(self.base_hmm.forward(t_skip=5) > -2000.0)

    def test_multi_train(self):
        """ Test training
        """
        log_like = self.base_hmm.multi_train(self.y,
                                             n_iterations=10,
                                             display=False)
        # Check that log likelihood increases montonically
        for i in range(1, len(log_like)):
            self.assertTrue(
                log_like[i - 1] < log_like[i] + 1e-14)  # Todo: fudge?
        # Check that trained model is close to true model
        numpy.testing.assert_allclose(
            self.base_hmm.y_mod.underlying_model._py_state.values(),
            self._py_state,
            atol=0.15)
        numpy.testing.assert_allclose(self.base_hmm.p_state2state.values(),
                                      self.p_state2state,
                                      atol=0.15)


class TestObservation_with_bundles(BaseClass):
    """ Test hmm.base.Observation_with_bundles
    """

    def setUp(self):
        self.y_class = hmm.base.IntegerObservation
        self.p_ys = hmm.simple.Prob(self._py_state.copy())
        self.rng = numpy.random.default_rng(0)

        self.Observation_with_bundles = hmm.base.Observation_with_bundles(
            self.y_class(self.p_ys, self.rng), self.bundle2state, self.rng)

        # outs[t] = (bundle[t], y[t])
        outs = [
            self.Observation_with_bundles.random_out(s)
            for s in range(len(self.p_ys))
        ]

        bundles = [out[0] for out in outs]
        ys = [out[1] for out in outs]
        self.data = [hmm.base.Bundle_segment(bundles, ys) for x in range(3)]

    def test_bundle_segment_str(self):
        self.assertTrue(
            str(self.data[0]) ==
            'y values:[1, 1, 2, 3, 5, 5]\nbundle values:[0, 0, 0, 1, 2, 2]\n')

    def test_bundle_segment_len(self):
        self.assertTrue(len(self.data[0]) == 6)

    def test_observe(self):
        t_seg = self.Observation_with_bundles.observe(self.data)
        numpy.testing.assert_equal(t_seg, numpy.array([0, 6, 12, 18]))

    def test_calculate(self):
        self.Observation_with_bundles.observe(self.data)
        result = self.Observation_with_bundles.calculate()
        self.assertTrue(result.min() == 0)
        self.assertTrue(result.max() > .35)

    def test_reestimate(self):
        self.Observation_with_bundles.observe(self.data)
        w = self.Observation_with_bundles.calculate()
        self.Observation_with_bundles.reestimate(w)
        result = self.Observation_with_bundles.underlying_model._py_state
        self.assertTrue(result.min() == 0)
        self.assertTrue(result.max() == 1.0)


# --------------------------------
# Local Variables:
# mode: python
# End:



"""observe_float.py: Various observation models for floating point measurements.

Use these models with the Hidden Markov Models in base.py.

"""
# pylint: disable = attribute-defined-outside-init

from __future__ import annotations  # Enables, eg, (self: HMM,

import typing

import numpy

import hmm.base

COPYRIGHT = """Copyright 2021 Andrew M. Fraser

This file is part of hmm.

Hmm is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

Hmm is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

See the file gpl.txt in the root directory of the hmm distribution
or see <http://www.gnu.org/licenses/>.
"""


class Gauss(hmm.base.IntegerObservation):
    r"""Scalar Gaussian observation model

    Args:
        mu: numpy.ndarray Means; a value for each state
        variance: numpy.ndarray Variances; a value for each state.
        rng: Generator with state

    The probability of observation y given state s is:
    Prob(y|s) = 1/\sqrt(2\pi var[s]) exp(-(y-mu[s])^2/(2*var[s])

    """
    _parameter_keys = "mu variance".split()

    def __init__(  # pylint: disable = super-init-not-called
        self: Gauss,
        mu: numpy.ndarray,
        variance: numpy.ndarray,
        rng: numpy.random.Generator,
    ):
        assert len(variance) == len(mu)
        self.mu = mu
        self.variance = variance
        self.sigma = numpy.sqrt(variance)
        self._rng = rng
        self.dtype = [numpy.float64]
        self.norm = 1 / numpy.sqrt(2 * numpy.pi * variance)
        self.n_states = len(variance)

    # Ignore: Super returned int
    def random_out(  # type: ignore
            # pylint: disable = arguments-differ
            self: Gauss,
            s: int) -> float:
        """ For simulation, draw a random observation given state s

        Args:
            s: Index of state

        Returns:
            Random observation drawn from distribution conditioned on state s

        """
        return (self._rng.normal(self.mu[s], self.sigma[s]))

    def calculate(self: Gauss,) -> numpy.ndarray:
        """Calculate and return likelihoods of states for all observations.

        Returns:
            self.p_y with self.p_y[t,s] = Prob(y[t]|state[t]=s)

        """
        assert self._y.reshape((-1, 1)).shape == (self.n_times, 1)
        d = self.mu - self._y.reshape((-1, 1))
        self._likelihood = numpy.exp(-d * d / (2 * self.variance)) * self.norm
        return self._likelihood

    # ToDo: Why do I need Any to make mypy happy?
    def diffs_to_var(self: Gauss, diffs: numpy.ndarray,
                     wsum: typing.Any) -> typing.Any:
        """ Formula for reestimating variance.

        Args:
            diffs
            wsum

        Returns:
            New estimate of variance

        Not in-line for convenience of MAP subclass
        """
        return (diffs * diffs).sum(axis=0) / wsum

    # Ignore: Super has optional argument warn
    def reestimate(  # type: ignore
            # pylint: disable = arguments-differ
            self: Gauss,
            w: numpy.ndarray):
        """
        Estimate new model parameters.  self._y already assigned

        Args:
            w: Weights; Prob(state[t]=s) given data and
                old model

        """
        assert len(self._y) > 0
        y = self._y
        wsum = w.sum(axis=0)
        self.mu = (w.T * y).sum(axis=1) / wsum
        diffs = (self.mu - y.reshape((-1, 1))) * numpy.sqrt(w)
        self.variance = self.diffs_to_var(diffs, wsum)
        self.sigma = numpy.sqrt(self.variance)
        self.norm = 1 / numpy.sqrt(2 * numpy.pi * self.variance)


class GaussMAP(Gauss):
    """For the variance use maximum a posteriori probability estimation
    instead of maximum likelihood.  The prior is inverse gamma

    """

    def __init__(self: GaussMAP,
                 mu: numpy.ndarray,
                 variance: numpy.ndarray,
                 rng: numpy.random.Generator,
                 alpha: float = 1,
                 beta: float = 1):
        super().__init__(mu, variance, rng)
        self.alpha = alpha
        self.beta = beta

    def log_prior(self):
        return_value = 0.0
        for s in range(self.n_states):
            return_value += -(self.alpha + 1) * numpy.log(
                self.variance[s]) - self.beta / self.variance[s]
        return return_value

    def diffs_to_var(self: GaussMAP, diffs: numpy.ndarray, wsum: float):
        numerator = 2 * self.beta + (diffs * diffs).sum(axis=0)
        denominator = 2 * self.alpha + 2 + wsum
        return numerator / denominator


class MultivariateGaussian(hmm.base.Observation_0):
    """Observation model for vector measurements.

    Args:
        mu[n_states, dimension]: Mean of distribution for each state
        sigma[n_states, dimension, dimension]: Covariance matrix for each state
        rng: Random number generator with state
        nu: Inverse Wishart parameter
        Psi: Inverse Wishart parameter
        small: Raise error if total likelihood of any observation is
            less than small

    Implements Maximum a posteriori probability estimation of the
    variance of each state.  Without data, sigma[s,i,i] =
    \frac{Psi}{nu + dim +1} where dim is the dimension of the
    observations

    """
    _parameter_keys = "mu sigma".split()

    def __init__(  # pylint: disable = super-init-not-called
        self: MultivariateGaussian,
        mu: numpy.ndarray,
        sigma: numpy.ndarray,
        rng: numpy.random.Generator,
        psi: float = .1,
        Psi: typing.Union[numpy.ndarray, None] = None,
        nu: float = 4,
        small=1.0e-100,
    ):
        # Check arguments
        self.n_states, self.dimension = mu.shape
        assert sigma.shape == (self.n_states, self.dimension, self.dimension)
        assert isinstance(rng, numpy.random.Generator)

        # Assign arguments to self
        self.mu = mu
        self.sigma = sigma
        self.inverse_sigma = numpy.empty(
            (self.n_states, self.dimension, self.dimension))
        self.norm = numpy.empty(self.n_states)
        for s in range(self.n_states):
            self.inverse_sigma[s, :, :] = numpy.linalg.inv(self.sigma[s, :, :])
            determinant = numpy.linalg.det(sigma[s, :, :])
            self.norm[s] = 1 / numpy.sqrt(
                (2 * numpy.pi)**self.dimension * determinant)
        self._rng = rng
        if Psi is None:
            Psi = numpy.eye(self.dimension) * psi
        self.Psi = Psi
        self.nu = nu
        self.small = small

    def random_out(  # pylint: disable = arguments-differ
            self: MultivariateGaussian, s: int) -> numpy.ndarray:
        return self._rng.multivariate_normal(self.mu[s], self.sigma[s])

    def __str__(self: MultivariateGaussian) -> str:
        save = numpy.get_printoptions()['precision']
        numpy.set_printoptions(precision=3)
        rv = 'Model %s instance\n' % self.__class__
        for s in range(self.n_states):
            rv += 'For state %d:\n' % s
            rv += ' inverse_sigma = \n%s\n' % self.inverse_sigma[s]
            rv += ' mu = %s' % self.mu[s]
            rv += ' norm = %f\n' % self.norm[s]
        numpy.set_printoptions(precision=save)
        return rv

    def calculate(self: MultivariateGaussian) -> numpy.ndarray:
        """
        Calculate and return likelihoods.

        Returns:
            self.p_y[t,s] = Prob(y[t]|state[t]=s)

        Assumes self.observe has assigned a single numpy.ndarray to self._y
        """
        assert self._y.shape == (
            self.n_times,
            self.dimension), 'You must call observe before calling calculate.'
        for t in range(self.n_times):
            for s in range(self.n_states):
                d = (self._y[t] - self.mu[s])
                dQd = float(numpy.dot(d, numpy.dot(self.inverse_sigma[s], d)))
                if dQd > 300:  # Underflow
                    self._likelihood[t, s] = 0
                else:
                    self._likelihood[t, s] = self.norm[s] * numpy.exp(-dQd / 2)
            if self._likelihood[t, :].sum() < self.small:
                raise ValueError("""Observation is not plausible from any state.
self.likelihood[{0},:]={1}""".format(t, self._likelihood[t, :]))
        return self._likelihood

    def reestimate(
        self: MultivariateGaussian,
        w: numpy.ndarray,
    ):
        """
        Estimate new model parameters.  self._y already assigned

        Args:
            w: Weights; Prob(state[t]=s) given data and
                old model

        """
        y = self._y
        wsum = w.sum(axis=0)
        self.mu = (numpy.inner(y.T, w.T) / wsum).T
        for s in range(self.n_states):
            rrsum = numpy.zeros((self.dimension, self.dimension))
            for t in range(self.n_times):
                r = y[t] - self.mu[s]
                rrsum += w[t, s] * numpy.outer(r, r)
            self.sigma[s, :, :] = (self.Psi + rrsum) / (wsum[s] + self.nu +
                                                        self.dimension + 1)
            det = numpy.linalg.det(self.sigma[s])
            assert (det > 0.0)
            self.inverse_sigma[s, :, :] = numpy.linalg.inv(self.sigma[s])
            self.norm[s] = 1.0 / (numpy.sqrt(
                (2 * numpy.pi)**self.dimension * det))

    def log_prior(self):
        return_value = 0.0
        for s in range(self.n_states):
            log_det = numpy.log(numpy.linalg.det(self.sigma[s]))
            trace = numpy.dot(self.Psi, self.inverse_sigma[s]).trace()
            return_value += -(self.nu + self.dimension +
                              1) * log_det / 2 - trace / 2
        return return_value


class AutoRegressive(GaussMAP):
    r"""Scalar autoregressive model with Gaussian residuals

    Args:
        ar_coefficients[n_states, ar_order]: Auto-regressive coefficients
        variance[n_states]: Residual variance for each state
        rng: Random number generator with state
        alpha: Part of prior for variance
        beta: Part of prior for variance
        small: Throw error if likelihood at any time is less than small

    Model: likelihood[t,s] = Normal(mu_{t,s}, var[s]) at _y[t]
           where mu_{t,s} = ar_coefficients[s] \cdot _y[t-n_ar:t] + offset[s]

           In method calculate, likelihoods are set to zero for t
           closer to starting segment boundaries than AR-order because
           not enough prior measurements exist.

    """
    _parameter_keys = "ar_coefficients offset variance".split()

    def __init__(  # pylint: disable = super-init-not-called
            self: AutoRegressive,
            ar_coefficients: numpy.ndarray,
            offset: numpy.ndarray,
            variance: numpy.ndarray,
            rng: numpy.random.Generator,
            alpha: float = 4.0,
            beta: float = 16.0,
            small: float = 1.0e-100):
        assert len(variance.shape) == 1
        assert len(offset.shape) == 1
        assert len(ar_coefficients.shape) == 2

        self.n_states, self.ar_order = ar_coefficients.shape

        assert offset.shape[0] == self.n_states
        assert variance.shape[0] == self.n_states
        assert isinstance(rng, numpy.random.Generator)

        # Store offset in self.ar_coefficients_offset for convenience in
        # both calculating likelihoods and in re-estimation
        self.ar_coefficients_offset = numpy.empty(
            (self.n_states, self.ar_order + 1))
        self.ar_coefficients_offset[:, :self.ar_order] = ar_coefficients
        self.ar_coefficients_offset[:, self.ar_order] = offset

        self.variance = variance
        self.norm = numpy.empty(self.n_states)
        for s in range(self.n_states):
            self.norm[s] = 1 / numpy.sqrt(2 * numpy.pi * self.variance[s])
        self._rng = rng
        self.alpha = alpha
        self.beta = beta
        self.small = small

    def initialize_out(self: AutoRegressive):
        """ Prepare for calls to self.random_out
        """
        self.history = numpy.zeros(self.ar_order + 1)
        self.history[-1] = 1.0

    # ToDo: Better way to have different arguments for method of
    # subclass without mypy or pylint complaints
    def random_out(  # type: ignore
            # pylint: disable = arguments-differ
            self: AutoRegressive,
            s: int) -> numpy.ndarray:
        mu = numpy.dot(self.history, self.ar_coefficients_offset[s])
        rv = self._rng.normal(mu, self.variance[s])
        self.history[:self.ar_order - 1] = self.history[1:self.ar_order]
        self.history[self.ar_order - 1] = rv
        return rv

    def __str__(self: AutoRegressive) -> str:
        save = numpy.get_printoptions()['precision']
        numpy.set_printoptions(precision=3)
        rv = 'Model %s instance\n' % type(self)
        for s in range(self.n_states):
            rv += 'For state %d:\n' % s
            rv += ' variance = \n%s\n' % self.variance[s]
            rv += ' ar_coefficients = %s' % self.ar_coefficients_offset[s, :-1]
            rv += ' offset = %s' % self.ar_coefficients_offset[s, -1]
            rv += ' norm = %f\n' % self.norm[s]
        numpy.set_printoptions(precision=save)
        return rv

    def _concatenate(self: AutoRegressive, y_segs: typing.Sequence) -> tuple:
        """Attach context to self and return the modified concatenated data
        and segment information.

        Args:
            y_segs: Independent measurement sequences.  Each sequence
            is a 1-d numpy array.

        Returns:
            (all_data, Segment boundaries)

        This method shortens each segment by ar_order.  That enables
        having a true context for each element of self._y

        self.context will be used in calculate and reestimate.
        After values get assigned, context[t, :-1] = previous
        observations, and context[t, -1] = 1.0

        """
        length = 0
        t_seg = [0]
        for seg in y_segs:
            length += len(seg) - self.ar_order
            t_seg.append(length)
        all_data = numpy.empty(length)
        self.context = numpy.ones((length, self.ar_order + 1))
        for i in range(len(t_seg) - 1):
            all_data[t_seg[i]:t_seg[i + 1]] = y_segs[i][self.ar_order:]
            for delta_t in range(0, t_seg[i + 1] - t_seg[i]):
                self.context[t_seg[i] +
                             delta_t, :-1] = y_segs[i][delta_t:delta_t +
                                                       self.ar_order]
                # The one in the last place of context gets multiplied
                # by the last elements of self.ar_coefficients_offset
                # in self.calculate().  It is an offset term.
        return all_data, t_seg

    def calculate(self: AutoRegressive) -> numpy.ndarray:
        """
        Calculate and return likelihoods.

        Returns:
            likelihood where likelihood[t,s] = Prob(y[t]|state[t]=s)

        """
        assert self._y.shape == (self.n_times,)

        for t in range(self.n_times):
            delta = self._y[t] - numpy.dot(self.ar_coefficients_offset,
                                           self.context[t])
            exponent = -delta * delta / (2 * self.variance)
            if exponent.min() < -300:  # Underflow
                for s in range(self.n_states):
                    if exponent[s] < -300:
                        self._likelihood[t, s] = 0.0
                    else:
                        self._likelihood[t, s] = self.norm[s] * numpy.exp(
                            -delta[s] * delta[s] / (2 * self.variance[s]))
            else:
                self._likelihood[t, :] = self.norm * numpy.exp(
                    -delta * delta / (2 * self.variance))

            if self._likelihood[t, :].sum() < self.small:
                raise ValueError("""Observation is not plausible from any state.
self.likelihood[{0},:]={1}""".format(t, self._likelihood[t, :]))
        return self._likelihood

    # mypy objects: "incompatible with supertype Integerobservation"
    def reestimate(  # type: ignore
            self: AutoRegressive,
            w: numpy.ndarray,
    ):
        """
        Estimate new model parameters.  self._y already assigned

        Args:
            w: Weights. w[t,s] = Prob(state[t]=s) given data and
                old model

        """
        mask = w >= self.small  # Small weights confuse the residual
        # calculation in least_squares()
        w2 = mask * w
        wsum = w2.sum(axis=0)
        w1 = numpy.sqrt(w2)  # n_times x n_states array of weights

        for s in range(self.n_states):
            w_y = w1[:, s] * self._y
            w_context = (w1[:, s] * self.context.T).T
            # pylint: disable = unused-variable
            fit, residuals, rank, singular_values = numpy.linalg.lstsq(
                w_context, w_y, rcond=None)
            assert rank == self.ar_order + 1
            self.ar_coefficients_offset[s, :] = fit
            delta = w_y - numpy.inner(w_context, fit)
            sum_squared_error = numpy.inner(delta, delta)
            numerator = 2 * self.beta + numpy.inner(delta, delta)
            denominator = 2 * self.alpha + 2 + wsum[s]
            self.variance[s] = numerator / denominator
            self.norm[s] = 1 / numpy.sqrt(2 * numpy.pi * self.variance[s])


# --------------------------------
# Local Variables:
# mode: python
# End:
